/* GDK - The GIMP Drawing Kit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#include "config.h"
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <ctype.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <sys/vt.h>
#include <sys/kd.h>
#include <errno.h>
#include <signal.h>
#include <ncurses.h>

#include <sys/socket.h>
#include <sys/un.h>

#include "gdk.h"
#include "gdkfb.h"

#include "gdkprivate-fb.h"
#include "gdkinternals.h"
#include "gdkfbmanager.h"

/* Private variable declarations
 */
static int gdk_initialized = 0;			    /* 1 if the library is initialized,
						     * 0 otherwise.
						     */

#ifdef G_ENABLE_DEBUG
static const GDebugKey gdk_debug_keys[] = {
  {"misc",	    GDK_DEBUG_MISC},
  {"events",	    GDK_DEBUG_EVENTS},
};

static const int gdk_ndebug_keys = sizeof(gdk_debug_keys)/sizeof(GDebugKey);

#endif /* G_ENABLE_DEBUG */

GdkArgDesc _gdk_windowing_args[] = {
  { NULL }
};

static const GScannerConfig     fb_modes_scanner_config =
{
  (
   " \t\n"
   )                    /* cset_skip_characters */,
  (
   G_CSET_a_2_z
   G_CSET_A_2_Z
   )                    /* cset_identifier_first */,
  (
   G_CSET_a_2_z
   "_-0123456789"
   G_CSET_A_2_Z
   )                    /* cset_identifier_nth */,
  ( "#\n" )             /* cpair_comment_single */,
  
  FALSE                  /* case_sensitive */,
  
  FALSE                 /* skip_comment_multi */,
  TRUE                  /* skip_comment_single */,
  FALSE                 /* scan_comment_multi */,
  TRUE                  /* scan_identifier */,
  TRUE                  /* scan_identifier_1char */,
  FALSE                 /* scan_identifier_NULL */,
  TRUE                  /* scan_symbols */,
  FALSE                 /* scan_binary */,
  FALSE                 /* scan_octal */,
  FALSE                  /* scan_float */,
  FALSE                  /* scan_hex */,
  FALSE                 /* scan_hex_dollar */,
  FALSE                 /* scan_string_sq */,
  TRUE                  /* scan_string_dq */,
  TRUE                  /* numbers_2_int */,
  FALSE                 /* int_2_float */,
  FALSE                 /* identifier_2_string */,
  TRUE                  /* char_2_token */,
  FALSE                 /* symbol_2_token */,
  FALSE                 /* scope_0_fallback */,
};

enum {
  FB_MODE,
  FB_ENDMODE,
  FB_GEOMETRY,
  FB_TIMINGS,
  FB_LACED,
  FB_HSYNC,
  FB_VSYNC,
  FB_CSYNC,
  FB_EXTSYNC,
  FB_DOUBLE,
  FB_ACCEL
};

char *fb_modes_keywords[] =
{
  "mode",
  "endmode",
  "geometry",
  "timings",
  "laced",
  "hsync",
  "vsync",
  "csync",
  "extsync",
  "double",
  "accel"
};

static int
gdk_fb_set_mode (GdkFBDisplay *display)
{

  initscr();                    /* ncurses init */
  clear();
  refresh();
	  
  if (has_colors()) {           /* Terminal supports color? */
    start_color();
  
/* od Pavla  */
    if ((COLORS < 8) || (COLOR_PAIRS < 64)) {
      fprintf( stderr, "fatal: colors supported but not enough of them\n" );
      exit(5);
    }
    {
      int i, j;
      for (i = 0; i < 8; i++)
        for (j = 0; j < 8; j++) {
	  if (i==0 && j==0)	/* COLOR_BLACK, COLOR_BLACK */
	    continue;
          init_pair(i * 8 + j, j, i);   /* j = foreground, i = background */
        }
    }
  }

  return 0;
}

static GdkFBDisplay *
gdk_fb_display_new ()
{
  GdkFBDisplay *display;
  gint x;
  const gchar *helpstring;

  display = g_new0 (GdkFBDisplay, 1);
  
  if (gdk_fb_set_mode (display) < 0)
    {
      g_free (display);
      return NULL;
    }

  display->fb_width = COLS*PAVEL_MAGIC_X;
  display->fb_height = (LINES-1)*PAVEL_MAGIC_Y;
  display->fb_mem = g_malloc(display->fb_height * display->fb_width);
  display->fb_mem_cursed = g_malloc0(((display->fb_height+PAVEL_MAGIC_Y-1)/PAVEL_MAGIC_Y)
                                    *((display->fb_width +PAVEL_MAGIC_X-1)/PAVEL_MAGIC_X)
				    *sizeof(*display->fb_mem_cursed));

  for (x=0; x<COLS; x++)
    {
      color_set ((8*COLOR_RED /* bg */ + COLOR_GREEN /* fg */), NULL);
      mvaddch (LINES-1, x, ' ');
    }
  helpstring="CursedGTK+  Alt+TAB=Switch window  Shift-F4=Close window  Shift-F1=Redraw screen";
  mvaddstr(LINES-1,(COLS-strlen(helpstring))/2,helpstring);

  return display;
}

static void
gdk_fb_display_destroy (GdkFBDisplay *display)
{
  refresh();
  endwin();
  
  g_free (display);
}

void
_gdk_windowing_init (int *argc, char ***argv)
{
  if (gdk_initialized)
    return;

  /* Create new session and become session leader */
  setsid();

  gdk_display = gdk_fb_display_new ();

  if (!gdk_display)
    return;

  if (!gdk_fb_keyboard_init (TRUE))
    {
      g_warning ("Failed to initialize keyboard");
      gdk_fb_display_destroy (gdk_display);
      gdk_display = NULL;
      return;
    }
  
  if (!gdk_fb_mouse_init (TRUE))
    {
      g_warning ("Failed to initialize mouse");
      return;
    }

  /* Although atexit is evil, we need it here because otherwise the
   * keyboard is left in a bad state. you can still run 'reset' but
   * that gets annoying after running testgtk for the 20th time.
   */
  g_atexit(_gdk_windowing_exit);

  gdk_initialized = TRUE;

  _gdk_selection_property = gdk_atom_intern ("GDK_SELECTION", FALSE);

}

void
_gdk_windowing_set_default_display (GdkDisplay *display)
{
}

/*
 *--------------------------------------------------------------
 * gdk_pointer_grab
 *
 *   Grabs the pointer to a specific window
 *
 * Arguments:
 *   "window" is the window which will receive the grab
 *   "owner_events" specifies whether events will be reported as is,
 *     or relative to "window"
 *   "event_mask" masks only interesting events
 *   "confine_to" limits the cursor movement to the specified window
 *   "cursor" changes the cursor for the duration of the grab
 *   "time" specifies the time
 *
 * Results:
 *
 * Side effects:
 *   requires a corresponding call to gdk_pointer_ungrab
 *
 *--------------------------------------------------------------
 */

GdkGrabStatus
gdk_pointer_grab (GdkWindow *	  window,
		  gint		  owner_events,
		  GdkEventMask	  event_mask,
		  GdkWindow *	  confine_to,
		  GdkCursor *	  cursor,
		  guint32	  time)
{
  return gdk_fb_pointer_grab (window,
			      owner_events,
			      event_mask,
			      confine_to,
			      cursor,
			      time, FALSE);
}

static gboolean _gdk_fb_pointer_implicit_grab = FALSE;

GdkGrabStatus
gdk_fb_pointer_grab (GdkWindow *	  window,
		     gint		  owner_events,
		     GdkEventMask	  event_mask,
		     GdkWindow *	  confine_to,
		     GdkCursor *	  cursor,
		     guint32	          time,
		     gboolean             implicit_grab)
{
  g_return_val_if_fail (window != NULL, 0);
  g_return_val_if_fail (GDK_IS_WINDOW (window), 0);
  g_return_val_if_fail (confine_to == NULL || GDK_IS_WINDOW (confine_to), 0);

  if (_gdk_fb_pointer_grab_window)
    {
      if (implicit_grab && !_gdk_fb_pointer_implicit_grab)
	return GDK_GRAB_ALREADY_GRABBED;

      gdk_pointer_ungrab (time);
    }

  gdk_fb_window_send_crossing_events (NULL,
				      window,
				      GDK_CROSSING_GRAB);  

  if (event_mask & GDK_BUTTON_MOTION_MASK)
      event_mask |=
	GDK_BUTTON1_MOTION_MASK |
	GDK_BUTTON2_MOTION_MASK |
	GDK_BUTTON3_MOTION_MASK;
  
  _gdk_fb_pointer_implicit_grab = implicit_grab;

  _gdk_fb_pointer_grab_window = gdk_window_ref (window);
  _gdk_fb_pointer_grab_owner_events = owner_events;

  _gdk_fb_pointer_grab_confine = confine_to ? gdk_window_ref (confine_to) : NULL;
  _gdk_fb_pointer_grab_events = event_mask;
  _gdk_fb_pointer_grab_cursor = cursor ? gdk_cursor_ref (cursor) : NULL;


  
  if (cursor)
    gdk_fb_cursor_reset ();

  return GDK_GRAB_SUCCESS;
}

/*
 *--------------------------------------------------------------
 * gdk_display_pointer_ungrab
 *
 *   Releases any pointer grab
 *
 * Arguments:
 *
 * Results:
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

void
gdk_display_pointer_ungrab (GdkDisplay *display,
			    guint32     time)
{
  gdk_fb_pointer_ungrab (time, FALSE);
}

void
gdk_fb_pointer_ungrab (guint32 time, gboolean implicit_grab)
{
  gboolean have_grab_cursor = _gdk_fb_pointer_grab_cursor && 1;
  GdkWindow *mousewin;
  GdkWindow *old_grab_window;
  
  if (!_gdk_fb_pointer_grab_window)
    return;

  if (implicit_grab && !_gdk_fb_pointer_implicit_grab)
    return;

  if (_gdk_fb_pointer_grab_confine)
    gdk_window_unref (_gdk_fb_pointer_grab_confine);
  _gdk_fb_pointer_grab_confine = NULL;

  if (_gdk_fb_pointer_grab_cursor)
    gdk_cursor_unref (_gdk_fb_pointer_grab_cursor);
  _gdk_fb_pointer_grab_cursor = NULL;

  if (have_grab_cursor)
    gdk_fb_cursor_reset ();

  old_grab_window = _gdk_fb_pointer_grab_window;
  
  _gdk_fb_pointer_grab_window = NULL;

  _gdk_fb_pointer_implicit_grab = FALSE;

  mousewin = gdk_window_at_pointer (NULL, NULL);
  gdk_fb_window_send_crossing_events (old_grab_window,
				      mousewin,
				      GDK_CROSSING_UNGRAB);
  
  if (old_grab_window)
    gdk_window_unref (old_grab_window);
}

/*
 *--------------------------------------------------------------
 * gdk_display_pointer_is_grabbed
 *
 *   Tell wether there is an active x pointer grab in effect
 *
 * Arguments:
 *
 * Results:
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

gint
gdk_display_pointer_is_grabbed (GdkDisplay *display)
{
  return _gdk_fb_pointer_grab_window != NULL;
}

/*
 *--------------------------------------------------------------
 * gdk_keyboard_grab
 *
 *   Grabs the keyboard to a specific window
 *
 * Arguments:
 *   "window" is the window which will receive the grab
 *   "owner_events" specifies whether events will be reported as is,
 *     or relative to "window"
 *   "time" specifies the time
 *
 * Results:
 *
 * Side effects:
 *   requires a corresponding call to gdk_keyboard_ungrab
 *
 *--------------------------------------------------------------
 */

GdkGrabStatus
gdk_keyboard_grab (GdkWindow  *window,
		   gint        owner_events,
		   guint32     time)
{
  g_return_val_if_fail (window != NULL, 0);
  g_return_val_if_fail (GDK_IS_WINDOW (window), 0);

  if (_gdk_fb_pointer_grab_window)
    gdk_keyboard_ungrab (time);

  _gdk_fb_keyboard_grab_window = gdk_window_ref (window);
  _gdk_fb_keyboard_grab_owner_events = owner_events;
  
  return GDK_GRAB_SUCCESS;
}

/*
 *--------------------------------------------------------------
 * gdk_display_keyboard_ungrab
 *
 *   Releases any keyboard grab
 *
 * Arguments:
 *
 * Results:
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

void
gdk_display_keyboard_ungrab (GdkDisplay *display,
			     guint32     time)
{
  if (_gdk_fb_keyboard_grab_window)
    gdk_window_unref (_gdk_fb_keyboard_grab_window);
  _gdk_fb_keyboard_grab_window = NULL;
}

gboolean
gdk_pointer_grab_info_libgtk_only (GdkDisplay *display,
				   GdkWindow **grab_window,
				   gboolean   *owner_events)
{
  if (_gdk_fb_pointer_grab_window)
    {
      if (grab_window)
        *grab_window = (GdkWindow *)_gdk_fb_pointer_grab_window;
      if (owner_events)
        *owner_events = _gdk_fb_pointer_grab_owner_events;
      
      return TRUE;
    }
  else
    return FALSE;
}

gboolean
gdk_keyboard_grab_info_libgtk_only (GdkDisplay *display,
				    GdkWindow **grab_window,
				    gboolean   *owner_events)
{
  if (_gdk_fb_keyboard_grab_window)
    {
      if (grab_window)
        *grab_window = (GdkWindow *)_gdk_fb_keyboard_grab_window;
      if (owner_events)
        *owner_events = _gdk_fb_keyboard_grab_owner_events;

      return TRUE;
    }
  else
    return FALSE;
}


/*
 *--------------------------------------------------------------
 * gdk_screen_get_width
 *
 *   Return the width of the screen.
 *
 * Arguments:
 *
 * Results:
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

gint
gdk_screen_get_width (GdkScreen *screen)
{
  return gdk_display->fb_width;
}

/*
 *--------------------------------------------------------------
 * gdk_screen_get_height
 *
 *   Return the height of the screen.
 *
 * Arguments:
 *
 * Results:
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

gint
gdk_screen_get_height (GdkScreen *screen)
{
  return gdk_display->fb_height;
}

/*
 *--------------------------------------------------------------
 * gdk_screen_get_width_mm
 *
 *   Return the width of the screen in millimeters.
 *
 * Arguments:
 *
 * Results:
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

gint
gdk_screen_get_width_mm (GdkScreen *screen)
{
  return 0.5 + gdk_screen_width () * (25.4 / 72.);
}

/*
 *--------------------------------------------------------------
 * gdk_screen_get_height_mm
 *
 *   Return the height of the screen in millimeters.
 *
 * Arguments:
 *
 * Results:
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

gint
gdk_screen_get_height_mm (GdkScreen *screen)
{
  return 0.5 + gdk_screen_height () * (25.4 / 72.);
}

void
_gdk_windowing_display_set_sm_client_id (GdkDisplay*  display,
					 const gchar* sm_client_id)
{
}

extern void keyboard_shutdown(void);

void
_gdk_windowing_exit (void)
{
  struct sigaction action;

  /* don't get interrupted while exiting
   * (cf. gdkrender-fb.c:gdk_shadow_fb_init) */
  action.sa_handler = SIG_IGN;
  sigemptyset (&action.sa_mask);
  action.sa_flags = 0;

  gdk_fb_mouse_close ();
  /*leak  g_free (gdk_fb_mouse);*/
  
  gdk_fb_keyboard_close ();
  /*leak g_free (gdk_fb_keyboard);*/
  
  gdk_fb_display_destroy (gdk_display);
  
  gdk_display = NULL;
}

gchar *
gdk_get_display(void)
{
  return g_strdup ("/dev/tty");
}

void
gdk_display_beep (GdkDisplay *display)
{
  beep();
  flash();
}

/* utils */
static const guint type_masks[] = {
  GDK_SUBSTRUCTURE_MASK, /* GDK_DELETE		= 0, */
  GDK_STRUCTURE_MASK, /* GDK_DESTROY		= 1, */
  GDK_EXPOSURE_MASK, /* GDK_EXPOSE		= 2, */
  GDK_POINTER_MOTION_MASK, /* GDK_MOTION_NOTIFY	= 3, */
  GDK_BUTTON_PRESS_MASK, /* GDK_BUTTON_PRESS	= 4, */
  GDK_BUTTON_PRESS_MASK, /* GDK_2BUTTON_PRESS	= 5, */
  GDK_BUTTON_PRESS_MASK, /* GDK_3BUTTON_PRESS	= 6, */
  GDK_BUTTON_RELEASE_MASK, /* GDK_BUTTON_RELEASE	= 7, */
  GDK_KEY_PRESS_MASK, /* GDK_KEY_PRESS	= 8, */
  GDK_KEY_RELEASE_MASK, /* GDK_KEY_RELEASE	= 9, */
  GDK_ENTER_NOTIFY_MASK, /* GDK_ENTER_NOTIFY	= 10, */
  GDK_LEAVE_NOTIFY_MASK, /* GDK_LEAVE_NOTIFY	= 11, */
  GDK_FOCUS_CHANGE_MASK, /* GDK_FOCUS_CHANGE	= 12, */
  GDK_STRUCTURE_MASK, /* GDK_CONFIGURE		= 13, */
  GDK_VISIBILITY_NOTIFY_MASK, /* GDK_MAP		= 14, */
  GDK_VISIBILITY_NOTIFY_MASK, /* GDK_UNMAP		= 15, */
  GDK_PROPERTY_CHANGE_MASK, /* GDK_PROPERTY_NOTIFY	= 16, */
  GDK_PROPERTY_CHANGE_MASK, /* GDK_SELECTION_CLEAR	= 17, */
  GDK_PROPERTY_CHANGE_MASK, /* GDK_SELECTION_REQUEST = 18, */
  GDK_PROPERTY_CHANGE_MASK, /* GDK_SELECTION_NOTIFY	= 19, */
  GDK_PROXIMITY_IN_MASK, /* GDK_PROXIMITY_IN	= 20, */
  GDK_PROXIMITY_OUT_MASK, /* GDK_PROXIMITY_OUT	= 21, */
  GDK_ALL_EVENTS_MASK, /* GDK_DRAG_ENTER        = 22, */
  GDK_ALL_EVENTS_MASK, /* GDK_DRAG_LEAVE        = 23, */
  GDK_ALL_EVENTS_MASK, /* GDK_DRAG_MOTION       = 24, */
  GDK_ALL_EVENTS_MASK, /* GDK_DRAG_STATUS       = 25, */
  GDK_ALL_EVENTS_MASK, /* GDK_DROP_START        = 26, */
  GDK_ALL_EVENTS_MASK, /* GDK_DROP_FINISHED     = 27, */
  GDK_ALL_EVENTS_MASK, /* GDK_CLIENT_EVENT	= 28, */
  GDK_VISIBILITY_NOTIFY_MASK, /* GDK_VISIBILITY_NOTIFY = 29, */
  GDK_EXPOSURE_MASK, /* GDK_NO_EXPOSE		= 30, */
  GDK_SCROLL_MASK /* GDK_SCROLL            = 31 */
};

GdkWindow *
gdk_fb_other_event_window (GdkWindow *window,
			   GdkEventType type)
{
  guint32 evmask;
  GdkWindow *w;

  w = window;
  while (w != _gdk_parent_root)
    {
      /* Huge hack, so that we don't propagate events to GtkWindow->frame */
      if ((w != window) &&
	  (GDK_WINDOW_P (w)->window_type != GDK_WINDOW_CHILD) &&
	  (g_object_get_data (G_OBJECT (w), "gdk-window-child-handler")))
	  break;
	  
      evmask = GDK_WINDOW_OBJECT(window)->event_mask;

      if (evmask & type_masks[type])
	return w;
      
      w = gdk_window_get_parent (w);
    }
  
  return NULL;
}

GdkWindow *
gdk_fb_pointer_event_window (GdkWindow *window,
			     GdkEventType type)
{
  guint evmask;
  GdkModifierType mask;
  GdkWindow *w;
	  
  gdk_fb_mouse_get_info (NULL, NULL, &mask);
  
  if (_gdk_fb_pointer_grab_window &&
      !_gdk_fb_pointer_grab_owner_events)
    {
      evmask = _gdk_fb_pointer_grab_events;

      if (evmask & (GDK_BUTTON1_MOTION_MASK | GDK_BUTTON2_MOTION_MASK | GDK_BUTTON3_MOTION_MASK))
	{
	  if (((mask & GDK_BUTTON1_MASK) && (evmask & GDK_BUTTON1_MOTION_MASK)) ||
	      ((mask & GDK_BUTTON2_MASK) && (evmask & GDK_BUTTON2_MOTION_MASK)) ||
	      ((mask & GDK_BUTTON3_MASK) && (evmask & GDK_BUTTON3_MOTION_MASK)))
	    evmask |= GDK_POINTER_MOTION_MASK;
	}

      if (evmask & type_masks[type])
	return _gdk_fb_pointer_grab_window;
      else
	return NULL;
    }

  w = window;
  while (w != _gdk_parent_root)
    {
      /* Huge hack, so that we don't propagate events to GtkWindow->frame */
      if ((w != window) &&
	  (GDK_WINDOW_P (w)->window_type != GDK_WINDOW_CHILD) &&
	  (g_object_get_data (G_OBJECT (w), "gdk-window-child-handler")))
	  break;
      
      evmask = GDK_WINDOW_OBJECT(window)->event_mask;

      if (evmask & (GDK_BUTTON1_MOTION_MASK | GDK_BUTTON2_MOTION_MASK | GDK_BUTTON3_MOTION_MASK))
	{
	  if (((mask & GDK_BUTTON1_MASK) && (evmask & GDK_BUTTON1_MOTION_MASK)) ||
	      ((mask & GDK_BUTTON2_MASK) && (evmask & GDK_BUTTON2_MOTION_MASK)) ||
	      ((mask & GDK_BUTTON3_MASK) && (evmask & GDK_BUTTON3_MOTION_MASK)))
	    evmask |= GDK_POINTER_MOTION_MASK;
	}

      if (evmask & type_masks[type])
	return w;
      
      w = gdk_window_get_parent (w);
    }

  if (_gdk_fb_pointer_grab_window &&
      _gdk_fb_pointer_grab_owner_events)
    {
      evmask = _gdk_fb_pointer_grab_events;
      
      if (evmask & (GDK_BUTTON1_MOTION_MASK | GDK_BUTTON2_MOTION_MASK | GDK_BUTTON3_MOTION_MASK))
	{
	  if (((mask & GDK_BUTTON1_MASK) && (evmask & GDK_BUTTON1_MOTION_MASK)) ||
	      ((mask & GDK_BUTTON2_MASK) && (evmask & GDK_BUTTON2_MOTION_MASK)) ||
	      ((mask & GDK_BUTTON3_MASK) && (evmask & GDK_BUTTON3_MOTION_MASK)))
	    evmask |= GDK_POINTER_MOTION_MASK;
	}
      
      if (evmask & type_masks[type])
	return _gdk_fb_pointer_grab_window;
    }
  
  return NULL;
}

GdkWindow *
gdk_fb_keyboard_event_window (GdkWindow *window,
			      GdkEventType type)
{
  guint32 evmask;
  GdkWindow *w;
  
  if (_gdk_fb_keyboard_grab_window &&
      !_gdk_fb_keyboard_grab_owner_events)
    {
      return _gdk_fb_keyboard_grab_window;
    }
  
  w = window;
  while (w != _gdk_parent_root)
    {
      /* Huge hack, so that we don't propagate events to GtkWindow->frame */
      if ((w != window) &&
	  (GDK_WINDOW_P (w)->window_type != GDK_WINDOW_CHILD) &&
	  (g_object_get_data (G_OBJECT (w), "gdk-window-child-handler")))
	  break;
	  
      evmask = GDK_WINDOW_OBJECT(window)->event_mask;

      if (evmask & type_masks[type])
	return w;
      
      w = gdk_window_get_parent (w);
    }
  
  if (_gdk_fb_keyboard_grab_window &&
      _gdk_fb_keyboard_grab_owner_events)
    {
      return _gdk_fb_keyboard_grab_window;
    }

  return NULL;
}

GdkEvent *
gdk_event_make (GdkWindow *window,
		GdkEventType type,
		gboolean append_to_queue)
{
  GdkEvent *event = gdk_event_new (type);
  guint32 the_time;
  
  the_time = gdk_fb_get_time ();
  
  event->any.window = gdk_window_ref (window);
  event->any.send_event = FALSE;
  switch (type)
    {
    case GDK_MOTION_NOTIFY:
      event->motion.time = the_time;
      event->motion.axes = NULL;
      break;
    case GDK_BUTTON_PRESS:
    case GDK_2BUTTON_PRESS:
    case GDK_3BUTTON_PRESS:
    case GDK_BUTTON_RELEASE:
      event->button.time = the_time;
      event->button.axes = NULL;
      break;
    case GDK_KEY_PRESS:
    case GDK_KEY_RELEASE:
      event->key.time = the_time;
      break;
    case GDK_ENTER_NOTIFY:
    case GDK_LEAVE_NOTIFY:
      event->crossing.time = the_time;
      break;
      
    case GDK_PROPERTY_NOTIFY:
      event->property.time = the_time;
      break;
      
    case GDK_SELECTION_CLEAR:
    case GDK_SELECTION_REQUEST:
    case GDK_SELECTION_NOTIFY:
      event->selection.time = the_time;
      break;
    case GDK_PROXIMITY_IN:
    case GDK_PROXIMITY_OUT:
      event->proximity.time = the_time;
      break;
    case GDK_DRAG_ENTER:
    case GDK_DRAG_LEAVE:
    case GDK_DRAG_MOTION:
    case GDK_DRAG_STATUS:
    case GDK_DROP_START:
    case GDK_DROP_FINISHED:
      event->dnd.time = the_time;
      break;
      
    case GDK_FOCUS_CHANGE:
    case GDK_CONFIGURE:
    case GDK_MAP:
    case GDK_UNMAP:
    case GDK_CLIENT_EVENT:
    case GDK_VISIBILITY_NOTIFY:
    case GDK_NO_EXPOSE:
    case GDK_SCROLL:
    case GDK_DELETE:
    case GDK_DESTROY:
    case GDK_EXPOSE:
    default:
      break;
    }
  
  if (append_to_queue)
    _gdk_event_queue_append (gdk_display_get_default (), event);
  
  return event;
}

void
gdk_fb_set_rotation (GdkFBAngle angle)
{
  if (angle == _gdk_fb_screen_angle)
    return;
  g_warning ("Screen rotation not supported.");
}

void
gdk_error_trap_push (void)
{
}

gint
gdk_error_trap_pop (void)
{
  return 0;
}

void
gdk_notify_startup_complete (void)
{
}

/* Cursed specific - refresh */

static gboolean registracni_razitko = FALSE;

gboolean
gdk_obcerstvovac_obrazovky (void)
{
  refresh();
  registracni_razitko = FALSE;

  return FALSE;	/* remove-me */
}

void 
gdk_obchodni_rejstrik_obcerstvovny (void)
{
  if (registracni_razitko == FALSE) 
    {
      registracni_razitko = TRUE;
      g_idle_add_full(G_PRIORITY_LOW, (GSourceFunc)gdk_obcerstvovac_obrazovky, NULL, NULL);
    }     
}

