/* GDK - The GIMP Drawing Kit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gdk/gdk.h>
#include <gdk/gdkinternals.h>
#include "gdkprivate-fb.h"
#include "gdkinputprivate.h"
#include <stdlib.h>
#include <sys/types.h>
#include <ncurses.h>


/*linux-fb*/
typedef struct _GdkFBMouse GdkFBMouse;

struct _GdkFBMouse {
  gint fd; /* Set by open */

  gint x, y;
  gboolean button_pressed[5];

  gboolean click_grab;
  GIOChannel *io;
  gint io_tag;

};
static GdkFBMouse *gdk_fb_mouse = NULL;


#ifdef HAVE_GPM

#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include <linux/keyboard.h> /* to use KG_SHIFT and so on */

#define ALL_KEY_MODS ((1<<KG_SHIFT)|(1<<KG_ALT)|(1<<KG_ALTGR)|(1<<KG_CTRL))

#include <gpm.h>

static gboolean handle_mouse_io              (GIOChannel   *gioc,
                                              GIOCondition  cond,
                                              gpointer      data);
static void gdk_fb_mouse_fake_gpm_packet (GdkFBMouse   *mouse,
					  gboolean     *got_motion);
					     

static void
handle_mouse_movement(GdkFBMouse *mouse)
{
  GdkWindow *mousewin;
  GdkEvent *event;
  gint x, y;
  GdkWindow *old_win, *win, *event_win, *cursor_win;
  guint state;
  GdkDrawableFBData *mousewin_private;

  old_win = gdk_window_at_pointer (NULL, NULL);

  if (_gdk_fb_pointer_grab_confine)
    mousewin = _gdk_fb_pointer_grab_confine;
  else
    mousewin = _gdk_parent_root;

  mousewin_private = GDK_DRAWABLE_IMPL_FBDATA (mousewin);

  if (mouse->x < mousewin_private->llim_x)
    mouse->x = mousewin_private->llim_x;
  else if (mouse->x > mousewin_private->lim_x - 1)
    mouse->x = mousewin_private->lim_x - 1;

  if (mouse->y < mousewin_private->llim_y)
    mouse->y = mousewin_private->llim_y;
  else if (mouse->y > mousewin_private->lim_y - 1)
    mouse->y = mousewin_private->lim_y - 1;

  win = gdk_window_at_pointer (NULL, NULL);

  cursor_win = win;
  if (_gdk_fb_pointer_grab_window)
    {
      GdkWindow *w;

      cursor_win = _gdk_fb_pointer_grab_window;
      w = win;
      while (w != _gdk_parent_root)
        {
          if (w == _gdk_fb_pointer_grab_window)
            {
              cursor_win = win;
              break;
            }
          w = gdk_window_get_parent (w);
        }
    }

  gdk_fb_cursor_move (mouse->x, mouse->y, cursor_win);

  event_win = gdk_fb_pointer_event_window (win, GDK_MOTION_NOTIFY);

  if (event_win && (win == old_win))
    {
      /* Only send motion events in the same window */
      gdk_window_get_origin (event_win, &x, &y);
      x = mouse->x - x;
      y = mouse->y - y;

      state = (mouse->button_pressed[0]?GDK_BUTTON1_MASK:0) |
        (mouse->button_pressed[1]?GDK_BUTTON2_MASK:0) |
        (mouse->button_pressed[2]?GDK_BUTTON3_MASK:0) |
        (mouse->button_pressed[3]?GDK_BUTTON4_MASK:0) |
        (mouse->button_pressed[4]?GDK_BUTTON5_MASK:0) |
        gdk_fb_keyboard_modifiers ();

      event = gdk_event_make (event_win, GDK_MOTION_NOTIFY, TRUE);
      event->motion.x = x;
      event->motion.y = y;
      event->motion.state = state;
      event->motion.is_hint = FALSE;
      event->motion.device = _gdk_core_pointer;
      event->motion.x_root = mouse->x;
      event->motion.y_root = mouse->y;
    }

  gdk_fb_window_send_crossing_events (NULL, win, GDK_CROSSING_NORMAL);
}

static void
send_button_event (GdkFBMouse *mouse,
                   guint button,
                   gboolean press_event)
{
  GdkEvent *event;
  gint x, y, i;
  GdkWindow *mouse_win;
  GdkWindow *event_win;
  int nbuttons;


  mouse_win = gdk_window_at_pointer(NULL, NULL);
  event_win = gdk_fb_pointer_event_window (mouse_win,
                                           press_event ? GDK_BUTTON_PRESS : GDK_BUTTON_RELEASE);

  if (event_win)
    {
      event = gdk_event_make (event_win, press_event ? GDK_BUTTON_PRESS : GDK_BUTTON_RELEASE, FALSE);

      gdk_window_get_origin (event_win, &x, &y);
      x = mouse->x - x;
      y = mouse->y - y;

      event->button.x = x;
      event->button.y = y;
      event->button.button = button;
      event->button.state = (mouse->button_pressed[0]?GDK_BUTTON1_MASK:0) |
        (mouse->button_pressed[1] ? GDK_BUTTON2_MASK : 0) |
        (mouse->button_pressed[2] ? GDK_BUTTON3_MASK : 0) |
        (mouse->button_pressed[3] ? GDK_BUTTON4_MASK : 0) |
        (mouse->button_pressed[4] ? GDK_BUTTON5_MASK : 0) |
        (1 << (button + 8)) /* badhack */ |
        gdk_fb_keyboard_modifiers ();
      event->button.device = _gdk_core_pointer;
      event->button.x_root = mouse->x;
      event->button.y_root = mouse->y;

      _gdk_event_queue_append (gdk_display_get_default (), event);

      /* For double-clicks */
      if (press_event)
        _gdk_event_button_generate (gdk_display_get_default (), event);
    }

  nbuttons = 0;
  for (i=0;i<5;i++)
    if (mouse->button_pressed[i])
      nbuttons++;

  /* Handle implicit button grabs: */
  if (press_event && nbuttons == 1)
    {
      gdk_fb_pointer_grab (mouse_win, FALSE,
                           gdk_window_get_events (mouse_win),
                           NULL, NULL,
                           GDK_CURRENT_TIME, TRUE);
      mouse->click_grab = TRUE;
    }
  else if (!press_event && nbuttons == 0 && mouse->click_grab)
    {
      gdk_fb_pointer_ungrab (GDK_CURRENT_TIME, TRUE);
      mouse->click_grab = FALSE;
    }
}


gboolean
gdk_fb_mouse_open (void)
{
  Gpm_Connect conn;

  conn.eventMask = ~0;
  conn.defaultMask= ~GPM_HARD;
  conn.maxMod = ~0;
  conn.minMod = 0;

  if (Gpm_Open(&conn, 0) == -1) {
    g_warning("Can't open GPM mouse connection\n");
    return FALSE;
  } else if (-2 == gpm_fd) {
    g_warning("Use rmev to see gpm events in xterm or rxvt\n"); 
    return FALSE;
  } 

  gdk_fb_mouse->fd = gpm_fd;
 
  gdk_fb_mouse->io =
    g_io_channel_unix_new (gdk_fb_mouse->fd);
  gdk_fb_mouse->io_tag =
    g_io_add_watch (gdk_fb_mouse->io,
                    G_IO_IN|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
                    handle_mouse_io, gdk_fb_mouse);
  
  return TRUE;
}

void 
gdk_fb_mouse_close (void)
{
  while (Gpm_Close()); /* close all the stack */
  return;
}

static gboolean
handle_mouse_io (GIOChannel *gioc,
                 GIOCondition cond,
                 gpointer data)
{
  GdkFBMouse *mouse = (GdkFBMouse *)data;
  gboolean got_motion = FALSE;

  gdk_fb_mouse_fake_gpm_packet (mouse, &got_motion);
  
  if (got_motion)
    handle_mouse_movement (mouse);

  return TRUE;
}

static void
gdk_fb_mouse_fake_gpm_packet (GdkFBMouse   *mouse,          /* compatibility...*/
                              gboolean     *got_motion)
			     
{
  Gpm_Event evt;
  gboolean new_button1 = 0;
  gboolean new_button2 = 0;
  gboolean new_button3 = 0;

  if (Gpm_GetEvent(&evt) > 0) {

    if (!(evt.type & GPM_UP))
      {
        new_button1 = !!(evt.buttons & (GPM_B_LEFT));
        new_button2 = !!(evt.buttons & (GPM_B_MIDDLE));
        new_button3 = !!(evt.buttons & (GPM_B_RIGHT));
      }    
    else
      {
        /* GPM_UP specifies only the buttons RELEASED right now. */
        new_button1 = (mouse->button_pressed[0] && !(evt.buttons & (GPM_B_LEFT)));
        new_button2 = (mouse->button_pressed[1] && !(evt.buttons & (GPM_B_MIDDLE)));
        new_button3 = (mouse->button_pressed[2] && !(evt.buttons & (GPM_B_RIGHT)));
      }    

  if (*got_motion &&
      (new_button1 != mouse->button_pressed[0] ||
       new_button2 != mouse->button_pressed[1] ||
       new_button3 != mouse->button_pressed[2]))
    {
      /* If a mouse button state changes we need to get correct ordering with enter/leave events,
         so push those out via handle_mouse_input */
      *got_motion = FALSE;
      handle_mouse_movement (mouse);
    }

    if (new_button1 != mouse->button_pressed[0])
      {
        mouse->button_pressed[0] = new_button1;
        send_button_event (mouse, 1, new_button1);
      }

   if (new_button2 != mouse->button_pressed[1])
     {
       mouse->button_pressed[1] = new_button2;
       send_button_event (mouse, 2, new_button2);
     }

   if (new_button3 != mouse->button_pressed[2])
     {
       mouse->button_pressed[2] = new_button3;
       send_button_event (mouse, 3, new_button3);
     }

   /* Never calculate the cursor position from evt.dx or evt.dy as our position
    * would get out of sync with evt.x and evt.y;
    * cursor is drawn at evt.x and evt.y by GPM_DRAWPOINTER()
    * while GTK+ position is read from mouse->x and mouse->y.
    * Position is (1,1)-based by gpm unless libgpm variable 'gpm_zerobased' is set.
    */
   mouse->x = ((evt.x-1) * PAVEL_MAGIC_X);
   mouse->y = ((evt.y-1) * PAVEL_MAGIC_Y);

   if (1
       && mouse->x >= 0 && mouse->x < gdk_display->fb_width
       && mouse->y >= 0 && mouse->y < gdk_display->fb_height)
     {
       struct GdkFB_mem_cursed *ukazovatko;

       ukazovatko = &gdk_display->fb_mem_cursed[mouse->x/PAVEL_MAGIC_X
		     +(gdk_display->fb_width+PAVEL_MAGIC_X-1)/PAVEL_MAGIC_X*(mouse->y/PAVEL_MAGIC_Y)];
       mouse->x += ukazovatko->offset_x;
       mouse->y += ukazovatko->offset_y;
     }

   if (evt.dx || evt.dy)
     *got_motion = TRUE;

   GPM_DRAWPOINTER(&evt);
  } else {
    g_warning("Oops, Gpm_GetEvent()\n");  
  }
}

#endif


#ifndef HAVE_GPM

gboolean
gdk_fb_mouse_open (void)
{
  return FALSE;
}

void 
gdk_fb_mouse_close (void)
{
  return;
}

#endif


void
gdk_fb_mouse_get_info (gint *x,
		       gint *y,
		       GdkModifierType *mask)
{
  if (x)
    *x = gdk_fb_mouse->x;
  if (y)
    *y = gdk_fb_mouse->y;
  if (mask)
    *mask =
      (gdk_fb_mouse->button_pressed[0]?GDK_BUTTON1_MASK:0) |
      (gdk_fb_mouse->button_pressed[1]?GDK_BUTTON2_MASK:0) |
      (gdk_fb_mouse->button_pressed[2]?GDK_BUTTON3_MASK:0) |
      gdk_fb_keyboard_modifiers (); 
}


#define _VTE_CAP_ESC "\033"		/* Escape */
#define _VTE_CAP_CSI _VTE_CAP_ESC "["	/* Control Sequence Introducer */
#define TRACKING_ALL_MOTION 1003

static void
decset(int mode, gboolean value)
{
  printf(_VTE_CAP_CSI "?%d%c", mode, value ? 'h' : 'l');
  fflush(stdout);
}

static void
tracking_close (void)
{
  decset(TRACKING_ALL_MOTION, FALSE);
}


static gboolean
gdk_fb_mouse_KEY_MOUSE_open_init_xterm (void)
{
  g_atexit (tracking_close);
  decset(TRACKING_ALL_MOTION, TRUE);

  return FALSE;	/* remove-me */
}

static void
gdk_fb_mouse_KEY_MOUSE_open (void)
{
char *term;

  /* Docs say -1 should disable BUTTON?_*CLICKED resolution.
   * RedHat ncurses-5.3-4 has IMO a bug
   * in ncurses/base/lib_mouse.c/mouseinterval()/"if (maxclick >= 0)"
   * where value -1 will not set anything.
   */
  mouseinterval (0);
  mousemask (0
	     | BUTTON1_PRESSED | BUTTON1_RELEASED
	     | BUTTON2_PRESSED | BUTTON2_RELEASED
	     | BUTTON3_PRESSED | BUTTON3_RELEASED
	     | BUTTON4_PRESSED | BUTTON4_RELEASED
	     | REPORT_MOUSE_POSITION,
	     NULL);

  if ((term = getenv("TERM")) && !strncmp(term, "xterm", 5))
    {
      /* We must be after gdkmain-fb.c:gdk_obchodni_rejstrik_obcerstvovny()
       * otherwise some ncurses initialization strings during the first refresh
       * will reset our decset().
       */
      g_idle_add_full(G_PRIORITY_LOW + 100, (GSourceFunc)gdk_fb_mouse_KEY_MOUSE_open_init_xterm, NULL, NULL);
    }
}


gboolean
gdk_fb_mouse_init (gboolean open_dev)
{
  gdk_fb_mouse = g_new0 (GdkFBMouse, 1);
  gdk_fb_mouse->fd = -1;
    
  gdk_fb_mouse->x = gdk_display->fb_width / 2;
  gdk_fb_mouse->y = gdk_display->fb_height / 2;
  
  if (open_dev)
    {
      gdk_fb_mouse_KEY_MOUSE_open ();
      gdk_fb_mouse_open ();
      return TRUE;	/* at least KEY_MOUSE should work */
    }
  else
    return TRUE;
}


void
gdk_fb_mouse_KEY_MOUSE_got (void)
{
  MEVENT mevent;
  gboolean new_button1 = 0;
  gboolean new_button2 = 0;
  gboolean new_button3 = 0;
  gboolean new_button4 = 0;
  gint new_x, new_y;
  gboolean got_motion;

  if (OK != getmouse (&mevent))
    return;

  new_button1 = ((gdk_fb_mouse->button_pressed[0] && !(mevent.bstate & BUTTON1_RELEASED))
		 || mevent.bstate & BUTTON1_PRESSED);
  new_button2 = ((gdk_fb_mouse->button_pressed[1] && !(mevent.bstate & BUTTON2_RELEASED))
		 || mevent.bstate & BUTTON2_PRESSED);
  new_button3 = ((gdk_fb_mouse->button_pressed[2] && !(mevent.bstate & BUTTON3_RELEASED))
		 || mevent.bstate & BUTTON3_PRESSED);
  new_button4 = ((gdk_fb_mouse->button_pressed[3] && !(mevent.bstate & BUTTON4_RELEASED))
		 || mevent.bstate & BUTTON4_PRESSED);

  new_x = (mevent.x * PAVEL_MAGIC_X);
  new_y = (mevent.y * PAVEL_MAGIC_Y);

  if (1
      && new_x >= 0 && new_x < gdk_display->fb_width
      && new_y >= 0 && new_y < gdk_display->fb_height)
    {
      struct GdkFB_mem_cursed *ukazovatko;

      ukazovatko = &gdk_display->fb_mem_cursed[new_x/PAVEL_MAGIC_X
		    +(gdk_display->fb_width+PAVEL_MAGIC_X-1)/PAVEL_MAGIC_X*(new_y/PAVEL_MAGIC_Y)];
      new_x += ukazovatko->offset_x;
      new_y += ukazovatko->offset_y;
    }

  if (new_button1 != gdk_fb_mouse->button_pressed[0])
    {
      gdk_fb_mouse->button_pressed[0] = new_button1;
      send_button_event (gdk_fb_mouse, 1, new_button1);
    }

  if (new_button2 != gdk_fb_mouse->button_pressed[1])
    {
      gdk_fb_mouse->button_pressed[1] = new_button2;
      send_button_event (gdk_fb_mouse, 2, new_button2);
    }

  if (new_button3 != gdk_fb_mouse->button_pressed[2])
    {
      gdk_fb_mouse->button_pressed[2] = new_button3;
      send_button_event (gdk_fb_mouse, 3, new_button3);
    }

  if (new_button4 != gdk_fb_mouse->button_pressed[3])
    {
      gdk_fb_mouse->button_pressed[3] = new_button4;
      send_button_event (gdk_fb_mouse, 4, new_button4);
    }

  got_motion = new_x != gdk_fb_mouse->x || new_y != gdk_fb_mouse->y;

  gdk_fb_mouse->x = new_x;
  gdk_fb_mouse->y = new_y;

  if (got_motion)
    handle_mouse_movement (gdk_fb_mouse);
}
